import numpy as np

from eradiate.experiments import AtmosphereExperiment
from eradiate.pipelines import AggregateRadiosity, GatherMono
from eradiate.units import symbol
from eradiate.units import unit_context_kernel as uck
from eradiate.units import unit_registry as ureg


def test_aggregate_radiosity(mode_mono):
    # Initialise test data
    irradiance = 2.0

    exp = AtmosphereExperiment(
        atmosphere=None,
        surface={"type": "lambertian", "reflectance": 1.0},
        illumination={"type": "directional", "irradiance": irradiance},
        measures=[
            {
                "type": "distant_flux",
                "film_resolution": (32, 32),
                "spp": 1000,
                "srf": {"type": "multi_delta", "wavelengths": [550.0] * ureg.nm},
            }
        ],
    )
    exp.process()
    values = GatherMono(
        var=("sector_radiosity", {"units": symbol(uck.get("irradiance"))})
    ).transform(exp.measures[0].mi_results)

    # Configure and apply step
    step = AggregateRadiosity(
        sector_radiosity_var="sector_radiosity", radiosity_var="radiosity"
    )
    result = step.transform(values)

    # Check that radiosity dimensions are correct
    assert not {"x_index", "y_index"}.issubset(result["radiosity"].dims)
    # This setup conserves energy
    assert np.isclose(irradiance, result["radiosity"], rtol=1e-4)



import numpy as np
import pytest

from eradiate import unit_registry as ureg
from eradiate.experiments import AtmosphereExperiment
from eradiate.pipelines import (
    AddIllumination,
    AddSpectralResponseFunction,
    AddViewingAngles,
    GatherMono,
    Pipeline,
)
from eradiate.pipelines._assemble import _remap_viewing_angles_plane
from eradiate.scenes.measure import HemisphericalDistantMeasure, MultiDistantMeasure
from eradiate.scenes.spectra import InterpolatedSpectrum, MultiDeltaSpectrum


@pytest.mark.parametrize(
    "illumination_type, expected_dims",
    [
        ("directional", ["sza", "saa"]),
        ("constant", []),
    ],
    ids=["directional", "constant"],
)
def test_add_illumination(modes_all_single, illumination_type, expected_dims):
    exp = AtmosphereExperiment(
        atmosphere=None,
        illumination={"type": illumination_type},
        measures=MultiDistantMeasure(
            srf=MultiDeltaSpectrum(wavelengths=[440.0, 550.0, 660.0] * ureg.nm)
        ),
    )
    exp.process()
    measure = exp.measures[0]

    # Apply basic post-processing
    steps = [("gather", GatherMono(var="radiance"))]

    values = Pipeline(steps=steps).transform(measure.mi_results)

    step = AddIllumination(illumination=exp.illumination, measure=measure)
    result = step.transform(values)
    # Irradiance is here and is indexed in Sun angles and spectral coordinate
    irradiance = result.data_vars["irradiance"]
    assert set(irradiance.dims) == {"w"}.union(set(expected_dims))


def test_remap_viewing_angles_plane():
    # Usage with "normal" parameters
    theta, phi = _remap_viewing_angles_plane(
        0 * ureg.deg,
        [60, 45, 0, 45, 60] * ureg.deg,
        [-180, -180, 0, 0, 0] * ureg.deg,
    )
    assert np.allclose(theta, [-60, -45, 0, 45, 60] * ureg.deg)
    assert np.allclose(phi, 0 * ureg.deg)

    # Usage with "exotic" parameters
    theta, phi = _remap_viewing_angles_plane(
        0 * ureg.deg,
        [60, 45, 0, 45, 60] * ureg.deg,
        [-180, 180, 90, 720, 0] * ureg.deg,
    )
    assert np.allclose(theta, [-60, -45, 0, 45, 60] * ureg.deg)
    assert np.allclose(phi, 0 * ureg.deg)

    # Improper direction ordering is detected
    with pytest.warns(Warning):
        theta, phi = _remap_viewing_angles_plane(
            180 * ureg.deg,
            [45, 0, 45] * ureg.deg,
            [-180, 0, 0] * ureg.deg,
        )
    assert np.allclose(theta, [45, 0, -45] * ureg.deg)
    assert np.allclose(phi, 180 * ureg.deg)


@pytest.mark.parametrize(
    "measure_type, expected_zenith, expected_azimuth",
    (
        (
            "multi_distant-aring",
            [45, 45, 45, 45, 45, 45, 45, 45],
            [0, 45, 90, 135, 180, 225, 270, 315],
        ),
        ("multi_distant-hplane", [-60, -45, 0, 45, 60], [0, 0, 0, 0, 0]),
        (
            "hemispherical_distant",
            [
                [41.409622, 41.409622],
                [41.409622, 41.409622],
            ],
            [
                [225, 135],
                [315, 45],
            ],
        ),
    ),
    ids=(
        "multi_distant-aring",
        "multi_distant-hplane",
        "hemispherical_distant",
    ),
)
def test_add_viewing_angles(mode_mono, measure_type, expected_zenith, expected_azimuth):
    # Initialise test data
    if measure_type == "multi_distant-aring":
        measure = MultiDistantMeasure.aring(
            zenith=45,
            azimuths=np.linspace(0.0, 360.0, 8, endpoint=False),
            spp=1,
        )

    elif measure_type == "multi_distant-hplane":
        measure = MultiDistantMeasure.hplane(
            zeniths=[-60, -45, 0, 45, 60],
            azimuth=0.0,
            spp=1,
        )

    elif measure_type == "hemispherical_distant":
        measure = HemisphericalDistantMeasure(film_resolution=(2, 2), spp=1)

    else:
        assert False

    exp = AtmosphereExperiment(atmosphere=None, measures=measure)
    exp.process()
    measure_index = 0
    measure = exp.measures[measure_index]

    # Apply basic post-processing
    values = GatherMono(var="radiance").transform(measure.mi_results)

    step = AddViewingAngles(measure=measure)
    result = step.transform(values)

    # Produced dataset has viewing angle coordinates
    assert "vza" in result.coords
    assert "vaa" in result.coords

    # Viewing angles are set to appropriate values
    assert np.allclose(result.coords["vza"].values.squeeze(), expected_zenith)
    assert np.allclose(result.coords["vaa"].values.squeeze(), expected_azimuth)


def test_add_srf(modes_all_single):
    exp = AtmosphereExperiment(
        atmosphere=None,
        measures=MultiDistantMeasure.hplane(
            zeniths=[-60, -45, 0, 45, 60],
            azimuth=0.0,
            spp=1,
            srf=InterpolatedSpectrum(
                wavelengths=[550.0, 560.0] * ureg.nm,
                values=[1.0, 1.0] * ureg.dimensionless,
            ),
        ),
    )
    exp.process()
    measure_index = 0
    measure = exp.measures[measure_index]

    # Apply basic post-processing
    steps = [GatherMono(var="radiance")]
    values = Pipeline(steps=steps).transform(measure.mi_results)

    step = AddSpectralResponseFunction(measure=measure)
    result = step.transform(values)

    # The spectral response function is added to the dataset as a data variable
    assert "srf" in result.data_vars
    # Its only dimension is wavelength
    assert set(result.srf.dims) == {"srf_w"}



"""
Molecular atmospheres.
"""

from __future__ import annotations

from copy import deepcopy

import attrs
import joseki
import numpy as np
import pint
import xarray as xr

import eradiate

from ._core import AbstractHeterogeneousAtmosphere
from ..core import traverse
from ..phase import PhaseFunction, RayleighPhaseFunction, phase_function_factory
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...converters import convert_absorption_data, convert_thermoprops
from ...exceptions import UnsupportedModeError
from ...radprops import (
    AtmosphereRadProfile,
    RadProfile,
    ZGrid,
)
from ...radprops._atmosphere import _absorption_data_repr
from ...radprops.absorption import DEFAULT_HANDLER_CONFIG
from ...spectral.ckd import BinSet, QuadSpec
from ...spectral.index import SpectralIndex
from ...spectral.mono import WavelengthSet
from ...units import unit_registry as ureg
from ...util.misc import summary_repr
from ...validators import validate_absorption_data


def default_absorption_data() -> tuple:
    """
    Default absorption data based on active spectral mode.

    Returns
    -------
    tuple
        Absorption data specifications.

    Raises
    ------
    UnsupportedModeError:
        When the spectral mode is neither 'mono' or 'ckd'.

    Notes
    -----
    The correct Eradiate mode must be set before this function is called.
    In monochromatic mode, the returned absorption data specifications are
    suitable for working in the 250 nm to 3125 nm wavelength range.
    In CKD mode, the returned absorption data specifications are suitable for
    working with the wavenumber band [18100, 18200] cm^-1 (equivalent to the
    wavelength range [549.45, 552.48] nm).
    The corresponding absorption datasets will be downloaded from the Eradiate
    online data store, if they have not already been downloaded.
    """
    wavelength_range = [549.5, 550.5] * ureg.nm
    if eradiate.mode().is_mono:
        dataset_codename = "komodo"
    elif eradiate.mode().is_ckd:
        dataset_codename = "monotropa"
    else:
        raise UnsupportedModeError(supported=["mono", "ckd"])
    return (dataset_codename, wavelength_range)


@parse_docs
@attrs.define(eq=False, slots=False)
class MolecularAtmosphere(AbstractHeterogeneousAtmosphere):
    """
    Molecular atmosphere scene element [``molecular``].

    See Also
    --------
    :class:`~eradiate.scenes.atmosphere.ParticleLayer`,
    :class:`~eradiate.scenes.atmosphere.HeterogeneousAtmosphere`.

    Notes
    -----
    This is commonly referred to as a clear-sky
    atmosphere, namely the atmosphere is free of clouds, aerosols or any other
    type of liquid or solid particles in suspension.

    It describes a gaseous mixture (air) whose thermophysical properties,
    namely pressure, temperature and constituent mole fractions, are allowed to
    vary with altitude (see the ``thermoprops`` attribute).

    The corresponding radiative properties are computed with this
    thermophysical profile as input.

    Special care must be taken that the absorption data is able to accomodate
    the specified thermophysical profile, especially along the constituent mole
    fractions axes (see also ``error_handler_config``).
    Note that the absoprtion data is able to support one spectral mode at a
    time. As a result, the Eradiate mode must be selected before instantiating
    this class, and the relevant absorption data must be provided.

    The Rayleigh scattering theory is used to compute the air volume
    scattering coefficient.

    The scattering phase function defaults to the Rayleigh scattering phase
    function but can be set to other values.
    """

    absorption_data: dict[tuple[pint.Quantity, pint.Quantity], xr.Dataset] = documented(
        attrs.field(
            kw_only=True,
            factory=default_absorption_data,
            converter=convert_absorption_data,
            validator=validate_absorption_data,
            repr=_absorption_data_repr,
        ),
        doc="Absorption coefficient data. "
        "If a file path, the absorption coefficient is loaded from the "
        "specified file (must be a NetCDF file). "
        "If a tuple, the first element is the dataset codename and the"
        "second is the desired working wavelength range.",
        type="dict[tuple[quantity, quantity], Dataset]",
        init_type="Dataset or list of Dataset or str or path-like or tuple[str, quantity]",
    )

    _thermoprops: xr.Dataset = documented(
        attrs.field(
            kw_only=True,
            factory=lambda: joseki.make(
                identifier="afgl_1986-us_standard",
                z=np.linspace(0.0, 120.0, 121) * ureg.km,
                additional_molecules=False,
            ),
            converter=convert_thermoprops,
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Thermophysical property dataset. If a path is passed, Eradiate will "
        "look it up and load it. If a dictionary is passed, it will be passed "
        "as keyword argument to ``joseki.make()``. The default is "
        '``joseki.make(identifier="afgl_1986-us_standard",  z=np.linspace(0.0, 120.0, 121) * ureg.km)``. '
        "See `the Joseki docs <https://rayference.github.io/joseki/latest/reference/#src.joseki.core.make>`_ "
        "for details.",
        type="Dataset",
        init_type="Dataset or path-like or dict",
    )

    _phase: PhaseFunction = documented(
        attrs.field(
            kw_only=True,
            factory=lambda: RayleighPhaseFunction(),
            converter=phase_function_factory.convert,
            validator=attrs.validators.instance_of(PhaseFunction),
        ),
        doc="Phase function.",
        type=":class:`.PhaseFunction`",
        init_type=":class:`.PhaseFunction` or dict",
        default=":class:`RayleighPhaseFunction() <.RayleighPhaseFunction>`",
    )

    has_absorption: bool = documented(
        attrs.field(
            kw_only=True,
            default=True,
            converter=bool,
        ),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(
            kw_only=True,
            default=True,
            converter=bool,
        ),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is set to zero.",
        type="bool",
        default="True",
    )

    @has_absorption.validator
    @has_scattering.validator
    def _switch_validator(self, attribute, value):
        if not self.has_absorption and not self.has_scattering:
            raise ValueError(
                f"while validating {attribute.name}: at least one of "
                "'has_absorption' and 'has_scattering' must be True"
            )

    _radprops_profile: AtmosphereRadProfile | None = attrs.field(
        default=None,
        validator=attrs.validators.optional(
            attrs.validators.instance_of(AtmosphereRadProfile)
        ),
        init=False,
        repr=False,
    )

    error_handler_config: dict[str, dict[str, str]] = documented(
        attrs.field(
            kw_only=True,
            factory=lambda: deepcopy(DEFAULT_HANDLER_CONFIG),
            validator=attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.deep_mapping(
                    key_validator=attrs.validators.instance_of(str),
                    value_validator=attrs.validators.instance_of(str),
                ),
            ),
        ),
        doc="Error handler configuration for absorption data interpolation.",
        type="dict",
        default=":data:`.DEFAULT_HANDLER_CONFIG`",
    )

    def update(self) -> None:
        # Inherit docstring
        self.phase.id = self.phase_id

        self.error_handler_config = {
            **DEFAULT_HANDLER_CONFIG,
            **self.error_handler_config,
        }

        self._radprops_profile = AtmosphereRadProfile(
            thermoprops=self.thermoprops,
            has_scattering=self.has_scattering,
            has_absorption=self.has_absorption,
            absorption_data=self.absorption_data,
            error_handler_config=self.error_handler_config,
        )

    def spectral_set(
        self, quad_spec: QuadSpec | None = None
    ) -> None | BinSet | WavelengthSet:
        if self.has_absorption:
            absorption_data = list(self.absorption_data.values())
            if len(absorption_data) == 1:
                absorption_data = absorption_data[0]

            if eradiate.mode().is_mono:
                return WavelengthSet.from_absorption_dataset(absorption_data)

            elif eradiate.mode().is_ckd:
                if quad_spec is None:
                    quad_spec = QuadSpec.default()  # default
                return BinSet.from_absorption_data(absorption_data, quad_spec)

            else:
                raise NotImplementedError
        else:
            return None

    # --------------------------------------------------------------------------
    #              Spatial extension and thermophysical properties
    # --------------------------------------------------------------------------

    @property
    def thermoprops(self) -> xr.Dataset:
        # Inherit docstring
        return self._thermoprops

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        min_sigma_s = self.radprops_profile.eval_sigma_s(ctx.si).min()
        return np.divide(
            1.0,
            min_sigma_s,
            where=min_sigma_s != 0.0,
            out=np.array([np.inf]),
        )

    # --------------------------------------------------------------------------
    #                             Radiative properties
    # --------------------------------------------------------------------------

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring
        return self._phase

    @property
    def radprops_profile(self) -> RadProfile:
        # Inherit docstring
        return self._radprops_profile

    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_albedo(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_t(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None, **kwargs
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_a(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_s(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    # --------------------------------------------------------------------------
    #                             Kernel dictionary
    # --------------------------------------------------------------------------

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _params_phase(self) -> dict:
        # Inherit docstring
        _, result = traverse(self.phase)
        return result.data



from __future__ import annotations

import logging

import attrs
import numpy as np
import xarray as xr
from pinttr.util import always_iterable

import eradiate

from ._core import PipelineStep
from ..attrs import documented, parse_docs
from ..cfconventions import ATTRIBUTES
from ..exceptions import UnsupportedModeError
from ..kernel import bitmap_to_dataset
from ..spectral.ckd import BinSet
from ..units import symbol
from ..units import unit_context_config as ucc

logger = logging.getLogger(__name__)


def _spectral_dims():
    if eradiate.mode().is_mono:
        return (("w", ATTRIBUTES["radiation_wavelength"]),)
    elif eradiate.mode().is_ckd:
        return (
            ("w", ATTRIBUTES["radiation_wavelength"]),
            ("g", ATTRIBUTES["quantile"]),
        )
    else:
        raise UnsupportedModeError


@parse_docs
@attrs.define
class GatherMono(PipelineStep):
    """
    Gather raw kernel results (output as nested dictionaries) into an xarray
    dataset.

    This pipeline step takes a nested dictionary produced by the parametric loop
    of an :class:`.Experiment` and repackages it as a :class:`~xarray.Dataset`.
    The top-level spectral index is mapped to mode-dependent spectral
    coordinates. Film dimensions are left unmodified and retain their
    metadata.

    An ``img`` variable holds sensor values. An ``spp`` variable holds the
    sample count.
    """

    var: str | tuple[str, dict] = documented(
        attrs.field(default="img"),
        default='"img"',
        type="str or tuple[str, dict]",
        init_type="str or tuple[str, dict], optional",
        doc="Name of the variable containing sensor data. Optionally, a "
        "(name, metadata) pair can be passed.",
    )

    def transform(self, x: dict) -> xr.Dataset:
        # Basic preparation
        spectral_dims = []
        spectral_dim_metadata = {}

        for y in _spectral_dims():
            if isinstance(y, str):
                spectral_dims.append(y)
                spectral_dim_metadata[y] = {}
            else:
                spectral_dims.append(y[0])
                spectral_dim_metadata[y[0]] = y[1]

        sensor_datasets = []

        # Loop on spectral indexes
        for siah, result_dict in x.items():
            if eradiate.mode().is_mono:
                spectral_index = siah
            elif eradiate.mode().is_ckd:
                spectral_index = siah

            ds = bitmap_to_dataset(result_dict["bitmap"])
            spp = result_dict["spp"]

            # Set spectral coordinates
            all_coords = {
                spectral_dim: [spectral_coord]
                for spectral_dim, spectral_coord in zip(
                    spectral_dims, always_iterable(spectral_index)
                )
            }

            # Add spectral and sensor dimensions to img array
            ds["img"] = ds.img.expand_dims(dim=all_coords)

            # Package spp in a data array
            all_dims = list(all_coords.keys())
            ds["spp"] = (all_dims, np.reshape(spp, [1 for _ in all_dims]))

            sensor_datasets.append(ds)

        # Combine all the data
        with xr.set_options(keep_attrs=True):
            result = xr.combine_by_coords(sensor_datasets)

        # Drop "channel" dimension when using a monochromatic Mitsuba variant
        if eradiate.mode().check(mi_color_mode="mono"):
            result = result.squeeze("channel", drop=True)

        for spectral_dim in spectral_dims:
            result[spectral_dim].attrs = spectral_dim_metadata[spectral_dim]

        # Apply metadata to data variables
        if isinstance(self.var, str):
            var = self.var
            var_metadata = {}
        else:
            var = self.var[0]
            var_metadata = self.var[1]

        result = result.rename({"img": var})
        result[var].attrs.update(var_metadata)

        return result


@parse_docs
@attrs.define
class GatherCKD(PipelineStep):
    """
    Gather raw kernel results into an xarray dataset.
    """

    binset: BinSet = documented(
        attrs.field(validator=attrs.validators.instance_of(BinSet)),
        doc="Bin set.",
        type=":class:`.BinSet`",
    )

    var: tuple[str, dict] = documented(
        attrs.field(default="img"),
        default='"img"',
        type="tuple[str, dict]",
        init_type="tuple[str, dict], optional",
        doc="Variable name containing sensor data and metadata.",
    )

    def transform(self, x: dict) -> xr.Dataset:
        # transform 'x' into a list of 'xr.Dataset' where each dataset
        # corresponds to a spectral index

        logger.debug("gather_ckd: begin")

        datasets = []

        # x is a dictionary whose keys are spectral indexes as hashable tuples
        # and whose values are dictionaries with keys "bitmap" and "spp"

        ix = 0
        bins = self.binset.bins
        # sort bins by wavelength
        bins = sorted(bins, key=lambda b: b.wcenter)
        items = list(x.items())

        for i, _bin in enumerate(bins):  # bin (w) loop
            ng = _bin.quad.weights.size
            _datasets = []

            for _ in range(ng):  # g loop
                item = items[ix]
                siah, result_dict = item
                w, g = siah  # wavelength, quantile pair
                bitmap = result_dict["bitmap"]
                spp = result_dict["spp"]
                dataset = bitmap_to_dataset(bitmap)

                # expand dimensions of 'img' data variable to include 'w' and 'g'
                dataset["img"] = dataset.img.expand_dims(dim={"w": [w], "g": [g]})

                # Drop "channel" dimension when using a monochromatic Mitsuba variant
                if eradiate.mode().check(mi_color_mode="mono"):
                    dataset = dataset.squeeze("channel", drop=True)

                # self.var is a tuple (name, metadata)
                name, metadata = self.var
                dataset = dataset.rename({"img": name})
                dataset[name].attrs.update(metadata)

                _datasets.append(dataset)
                ix += 1

            # concatenate along 'g'
            ds = xr.concat(_datasets, dim="g")

            # compute quadrature
            # this is a weighted sum array reduction:
            # https://docs.xarray.dev/en/stable/user-guide/computation.html#weighted-array-reductions

            # normalise weights to the [0, 1] g-interval
            weights_values = 0.5 * _bin.quad.weights
            weights = xr.DataArray(weights_values, dims=["g"], coords={"g": ds.g})
            with xr.set_options(keep_attrs=True):
                weighted = ds[name].weighted(weights)
                weighted_sum = weighted.sum(dim="g")

            #  create dataset for current bin (w)
            dataset = xr.Dataset()
            dataset[name] = weighted_sum

            # add 'spp' data variable
            dims = dataset.dims
            dataset["spp"] = (dims, spp * np.ones(tuple(dims.values())))

            # add 'wbounds' data variable, and 'wbv' coordinate
            wbounds = np.stack([_bin.wmin, _bin.wmax])
            wunits = ucc.get("wavelength")
            dataset["wbounds"] = (
                ["wbv", "w"],
                wbounds.m_as(wunits).reshape((2, 1)),
                {
                    "standard_name": "radiation_wavelength_bound",
                    "long_name": "wavelength bound",
                    "units": symbol(wunits),
                },
            )
            dataset["wbv"] = (["wbv"], ["lower", "upper"])

            datasets.append(dataset)

        with xr.set_options(keep_attrs=True):
            result = xr.concat(datasets, dim="w")

            # add metadata for 'w'
            result["w"].attrs.update(ATTRIBUTES["radiation_wavelength"])

        logger.debug("gather: end")
        return result



import mitsuba as mi
import numpy as np

from eradiate.pipelines import GatherMono


def test_pipeline_step_gather_mono(results_mono):
    """
    Basic test case for :class:`.Gather` (monochromatic mode).
    """
    values: mi.Bitmap = results_mono[0]

    # Configure this pipeline step according to the used sensor
    var = "radiance"
    step = GatherMono(var=var)
    result = step.transform(values)

    # Check that dimensions are those we expect
    spectral_dims = {"w": 1}
    film_dims = {
        "y_index": 32,
        "x_index": 32,
    }  # Note: the 'channel' film dimension is dropped in mono mode
    all_dims = {**spectral_dims, **film_dims}
    assert dict(result.dims) == all_dims

    # Check that data vars are those we expected
    assert set(result.keys()) == {"spp", var}
    # Sensor data has all dimensions
    assert set(result[var].dims) == set(all_dims.keys())

    # Check radiance values
    assert np.allclose(result[var].values, 2.0 / np.pi)


def test_pipeline_step_gather_ckd(results_ckd):
    """
    Basic test case for :class:`.Gather` (CKD).
    """
    values = results_ckd[0]

    # Configure this pipeline step according to the used sensor
    var = "radiance"
    step = GatherMono(var=var)
    result = step.transform(values)

    # Check that dimensions are those we expect
    spectral_dims = {"w": 1, "g": 1}  # 1 is the default value
    film_dims = {
        "y_index": 32,
        "x_index": 32,
    }  # Note: the 'channel' film dimension is dropped in mono mode
    all_dims = {**spectral_dims, **film_dims}
    assert dict(result.dims) == all_dims

    # Check that data vars are those we expected
    assert set(result.keys()) == {"spp", var}
    # Sensor data has all dimensions
    assert set(result[var].dims) == set(all_dims.keys())

    # Check radiance and sample count values
    assert np.allclose(result[var].values, 2.0 / np.pi)



from __future__ import annotations

import itertools
import logging
import typing as t
from abc import ABC, abstractmethod
from collections import Counter

import attrs

from ..attrs import documented, parse_docs
from ..util.misc import camel_to_snake

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
#                                Local utilities
# ------------------------------------------------------------------------------


def _pipeline_steps_converter(value):
    result = []

    for i, step in enumerate(value):
        if isinstance(step, PipelineStep):
            result.append((f"{i}_{camel_to_snake(step.__class__.__name__)}", step))
        else:
            result.append(step)

    return result


# ------------------------------------------------------------------------------
#                          Basic classes and interfaces
# ------------------------------------------------------------------------------


@attrs.define
class Pipeline:
    """
    A simple data processing pipeline remotely inspired from scikit-learn's
    ``Pipeline`` class.
    """

    steps: list[tuple[str, PipelineStep]] = attrs.field(
        factory=list, converter=_pipeline_steps_converter
    )

    @steps.validator
    def _steps_validator(self, attribute, value):
        for id, step in value:
            if not isinstance(step, PipelineStep):
                raise ValueError(
                    f"while validating '{attribute.name}': step '{id}' is not "
                    "a PipelineStep instance"
                )

        # Check for name duplicates
        for name, count in Counter([name for name, _ in value]).items():
            if count > 1:
                raise ValueError(
                    f"while validating '{attribute.name}': found duplicate "
                    f"step name {name}"
                )

    _names: list[str] = attrs.field(factory=list, init=False, repr=False)

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        """
        Update internal state. Should be run whenever `steps` is modified or
        mutated.
        """
        self._names = [name for name, _ in self.steps]

    def _step_index(self, name):
        return self._names.index(name)

    @property
    def named_steps(self) -> dict[str, PipelineStep]:
        """
        dict[str, :class:`.PipelineStep`]: A dictionary mapping names to their
            corresponding step.
        """
        return dict(self.steps)

    def add(
        self,
        name: str,
        step: PipelineStep,
        position: int | None = None,
        before: str | None = None,
        after: str | None = None,
    ) -> None:
        """
        Add a step to an existing pipeline.

        Parameters
        ----------
        name : str
            Name of the step to be added. If the passed name already exists, an
            exception will be raised.

        step : :class:`.PipelineStep`
            Step to be added to the pipeline.

        position : int, optional
             Index where `step` will be inserted.

        before : str, optional
            Insert `step` before the step with the name `name`. Exclusive with
            `after`.

        after : str, optional
            Insert `step` after the step with the name `name`. Exclusive with
            `before`.

        Raises
        ------
        ValueError
            If `name` maps to an already registered step.

        ValueError
            If both `before` and `after` are set.

        Notes
        -----
        * If none of `position`, `before` or `after` are set, the step will be
          appended to the pipeline.
        * If `position` and `before` (resp. `after`) are set, `before` (resp.
          `after`) takes precedence.
        """
        if name in self._names:
            raise ValueError(f"step name {name} already exists")

        if before is not None:
            if after is not None:
                raise ValueError("cannot set both 'before' and 'after' parameters")
            position = self._step_index(before)

        if after is not None:
            if before is not None:
                raise ValueError("cannot set both 'before' and 'after' parameters")
            position = self._step_index(after) + 1

        if position is None:
            self.steps.append((name, step))

        else:
            self.steps.insert(position, (name, step))

        self.update()

    def transform(
        self,
        x: t.Any,
        start: int | str | None = None,
        stop: int | str | None = None,
        stop_after: int | str | None = None,
        step: int | str | None = None,
    ) -> t.Any:
        """
        Apply the pipeline to a given data. Keyword arguments can be used to
        restrict pipeline execution to selected steps.

        Parameters
        ----------
        x
            Data to apply the pipeline to.

        start : int or str, optional
            If set, start execution at indexed step.

        stop : int or str, optional
            If set, stop execution at step preceding indexed step.

        stop_after : int or str, optional
            If set, stop execution after indexed step. Takes precedence over
            `stop`.

        step : int or str, optional
            If set, execute indexed step only. Takes precedence on all other
            step selectors.

        Returns
        -------
        xt
            Processed data.
        """
        if step is not None:
            if isinstance(step, str):
                step = self._step_index(step)

            return self.steps[step][1].transform(x)

        if isinstance(start, str):
            start = self._step_index(start)

        if isinstance(stop_after, str):
            stop_after = self._step_index(stop_after)

        if stop_after is not None:
            stop = stop_after + 1

        if isinstance(stop, str):
            stop = self._step_index(stop)

        xt = x
        for _, _, transform in self._iter(start, stop):
            xt = transform.transform(xt)
        return xt

    def _iter(self, start: int | None = None, stop: int | None = None):
        """
        Generate (idx, name, trans) tuples from self.steps.
        """
        if start is None:
            start = 0

        if stop is None:
            stop = len(self.steps)

        for idx, (name, trans) in enumerate(itertools.islice(self.steps, start, stop)):
            yield idx, name, trans

    @classmethod
    def convert(cls, x: t.Any) -> t.Any:
        if isinstance(x, (tuple, list)):
            return cls(x)

        else:
            return x


@attrs.define
class PipelineStep(ABC):
    """
    Interface for pipeline step definitions.
    """

    @abstractmethod
    def transform(self, x: t.Any) -> t.Any:
        """
        Apply the pipeline step to a given data.

        Parameters
        ----------
        x
            Data to process.

        Returns
        -------
        xt
            Processed data.
        """
        pass


@parse_docs
@attrs.define
class ApplyCallable(PipelineStep):
    """
    Turn a callable into a pipeline step.
    """

    callable: t.Callable = documented(
        attrs.field(validator=attrs.validators.is_callable()),
        type="callable",
        doc="Callable with signature ``f(x: Any) -> Any``.",
    )

    def transform(self, x: t.Any) -> t.Any:
        return self.callable(x)



from __future__ import annotations

import logging
import typing as t
import warnings

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

import eradiate

from ._core import PipelineStep
from ..attrs import documented, parse_docs
from ..exceptions import UnsupportedModeError
from ..frame import angles_in_hplane
from ..scenes.illumination import ConstantIllumination, DirectionalIllumination
from ..scenes.measure import Measure
from ..scenes.spectra import (
    InterpolatedSpectrum,
    Spectrum,
)
from ..units import symbol, to_quantity
from ..units import unit_context_kernel as uck
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)


@parse_docs
@attrs.define
class AddIllumination(PipelineStep):
    """
    Add illumination data.

    This post-processing pipeline step adds illumination data:

    * if `illumination` is a :class:`.DirectionalIllumination` instance, then
      a data variable (holding the incoming top-of-scene flux with respect to a
      horizontal surface) is created, with dimensions ``sza`` and ``vaa``;
    * if `illumination` is a :class:`.ConstantIllumination` instance, then
      the created data variable has no coordinate.
    """

    illumination: DirectionalIllumination | ConstantIllumination = documented(
        attrs.field(
            validator=attrs.validators.instance_of(
                (DirectionalIllumination, ConstantIllumination)
            ),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="An :class:`.Illumination` instance from which the illumination "
        "data originates.",
        type=":class:`.DirectionalIllumination` or :class:`.ConstantIllumination`",
    )

    measure: Measure = documented(
        attrs.field(
            validator=attrs.validators.instance_of(Measure),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="A :class:`.Measure` instance from the data originates.",
        type=":class:`.Measure`",
    )

    irradiance_var: str = documented(
        attrs.field(default="irradiance", validator=attrs.validators.instance_of(str)),
        doc="Name of the variable storing irradiance (incoming flux) values.",
        type="str",
        default='"irradiance"',
    )

    def transform(self, x: t.Any) -> t.Any:
        logger.debug("add_illumination: begin")
        k_irradiance_units = uck.get("irradiance")
        illumination = self.illumination
        # measure = self.measure
        result = x.copy(deep=False)

        # Collect spectral coordinate values for verification purposes
        wavelengths_dataset = to_quantity(x.w)

        def eval_illumination_spectrum(
            field_name: str, k_units: pint.Unit
        ) -> pint.Quantity:
            # Local helper function to help with illumination spectrum evaluation

            spectrum: Spectrum = getattr(illumination, field_name)
            results_wavelengths = to_quantity(x.w)

            if eradiate.mode().is_mono:
                wavelengths = results_wavelengths
                assert np.allclose(wavelengths, wavelengths_dataset)
                return spectrum.eval_mono(wavelengths).m_as(k_units)

            elif eradiate.mode().is_ckd:
                result = spectrum.eval_ckd(
                    w=results_wavelengths,
                    g=[0, 1],
                ).m_as(k_units)

                # Reorder data by ascending wavelengths
                # indices = results_wavelengths.argsort()
                wavelengths = results_wavelengths  # [indices]
                assert np.allclose(wavelengths, wavelengths_dataset)
                # result = result[indices]

                return result

            else:
                raise UnsupportedModeError(supported=("monochromatic", "ckd"))

        if isinstance(illumination, DirectionalIllumination):
            # Collect illumination angular data
            saa = illumination.azimuth.m_as(ureg.deg)
            sza = illumination.zenith.m_as(ureg.deg)
            cos_sza = np.cos(np.deg2rad(sza))

            # Add angular dimensions
            result = result.expand_dims({"sza": [sza], "saa": [saa]}, axis=(0, 1))
            result.coords["sza"].attrs = {
                "standard_name": "solar_zenith_angle",
                "long_name": "solar zenith angle",
                "units": symbol("deg"),
            }
            result.coords["saa"].attrs = {
                "standard_name": "solar_azimuth_angle",
                "long_name": "solar azimuth angle",
                "units": symbol("deg"),
            }

            # Collect illumination spectral data
            irradiances = eval_illumination_spectrum("irradiance", k_irradiance_units)

            # Add irradiance variable
            result[self.irradiance_var] = (
                ("sza", "saa", "w"),
                (irradiances * cos_sza).reshape((1, 1, len(irradiances))),
            )

        elif isinstance(illumination, ConstantIllumination):
            # Collect illumination spectral data
            k_radiance_units = uck.get("radiance")
            radiances = eval_illumination_spectrum("radiance", k_radiance_units)

            # Add irradiance variable
            result[self.irradiance_var] = (
                ("w",),
                np.pi * radiances.reshape((len(radiances),)),
            )

        else:
            raise TypeError(
                "field 'illumination' must be one of "
                "(DirectionalIllumination, ConstantIllumination), got a "
                f"{illumination.__class__.__name__}"
            )

        result[self.irradiance_var].attrs = {
            "standard_name": "horizontal_solar_irradiance_per_unit_wavelength",
            "long_name": "horizontal spectral irradiance",
            "units": symbol(k_irradiance_units),
        }
        logger.debug("add_illumination: end")
        return result


@parse_docs
@attrs.define
class AddViewingAngles(PipelineStep):
    """
    Create new ``vza`` and ``vaa`` coordinate variables mapping viewing angles
    to other coordinates.
    """

    measure: Measure = documented(
        attrs.field(
            validator=attrs.validators.instance_of(Measure),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="A :class:`.Measure` instance from which the processed data originates.",
        type=":class:`.Measure`",
    )

    def transform(self, x: t.Any) -> t.Any:
        measure = self.measure
        viewing_angles = measure.viewing_angles

        # Collect zenith and azimuth values
        theta = viewing_angles[:, :, 0]
        phi = viewing_angles[:, :, 1]

        # Attach coordinates
        with xr.set_options(keep_attrs=True):
            result = x.assign_coords(
                {
                    "vza": (
                        ("x_index", "y_index"),
                        theta.m_as(ureg.deg),
                        {
                            "standard_name": "viewing_zenith_angle",
                            "long_name": "viewing zenith angle",
                            "units": symbol("deg"),
                        },
                    ),
                    "vaa": (
                        ("x_index", "y_index"),
                        phi.m_as(ureg.deg),
                        {
                            "standard_name": "viewing_azimuth_angle",
                            "long_name": "viewing azimuth angle",
                            "units": symbol("deg"),
                        },
                    ),
                }
            )

        return result


@ureg.wraps(ret=("rad", "rad"), args=("rad", "rad", "rad"), strict=True)
def _remap_viewing_angles_plane(
    plane: np.typing.ArrayLike,
    theta: np.typing.ArrayLike,
    phi: np.typing.ArrayLike,
) -> tuple[np.typing.ArrayLike, np.typing.ArrayLike]:
    r"""
    Remap viewing angles to a hemisphere plane cut.

    Parameters
    ----------
    plane : quantity
         Plane cut orientation (scalar value).

    theta : quantity
        List of zenith angle values with (N,) shape.

    phi : quantity
        List of azimuth angle values with (N,) shape.

    Returns
    -------
    theta : quantity
        List of zenith angle values in :math:`[-\pi, \pi]` with (N,) shape.

    phi : quantity
        List of azimuth angle values in :math:`[0, 2\pi]` with (N,) shape
        (equal to `plane` modulo :math:`\pi`).

    Warns
    -----
    When zenith angle values are not sorted in ascending order.
    """
    # Normalize all angles
    twopi = 2.0 * np.pi
    plane = plane % twopi
    theta = theta % twopi
    phi = phi % twopi

    # Check that phi values are compatible with requested plane
    in_plane_positive, in_plane_negative = angles_in_hplane(
        plane, theta, phi, raise_exc=True
    )

    # Check if any point is allocated to both half-planes (uncomment to debug)
    # assert not np.any(in_plane_positive & in_plane_negative)

    # Normalise zenith values
    theta = np.where(in_plane_positive, theta, -theta)

    # Normalize azimuth values
    phi = np.full_like(theta, plane)

    # Check ordering and warn if it is not strictly increasing
    if not _is_sorted(theta):
        warnings.warn(
            "Viewing zenith angle values are not sorted in ascending order, "
            "you might want to consider changing direction definitions."
        )

    return theta, phi


def _is_sorted(x):
    return np.all(x[:-1] <= x[1:])


@parse_docs
@attrs.define
class AddSpectralResponseFunction(PipelineStep):
    """
    Add spectral response function data.

    This post-processing pipeline step adds spectral response function data to
    the processed dataset.
    """

    measure: Measure = documented(
        attrs.field(
            validator=attrs.validators.instance_of(Measure),
            repr=lambda self: f"{self.__class__.__name__}(id='{self.id}', ...)",
        ),
        doc="A :class:`.Measure` instance from which the processed data originates.",
        type=":class:`.Measure`",
    )

    def transform(self, x: t.Any) -> t.Any:
        logger.debug("add_spectral_response_function: begin")
        result = x.copy(deep=False)

        # Evaluate SRF
        srf = self.measure.srf

        if isinstance(srf, InterpolatedSpectrum):
            srf_w = srf.wavelengths
            srf_values = pinttr.util.ensure_units(srf.values, ureg.dimensionless)

        else:
            raise TypeError(f"unsupported SRF type '{srf.__class__.__name__}'")

        # Add SRF spectral coordinates to dataset
        w_units = ureg.nm
        result = result.assign_coords(
            {
                "srf_w": (
                    "srf_w",
                    srf_w.m_as(w_units),
                    {
                        "standard_name": "radiation_wavelength",
                        "long_name": "wavelength",
                        "units": "nm",
                    },
                )
            }
        )

        # Add SRF variable to dataset
        result["srf"] = ("srf_w", srf_values.m)
        result.srf.attrs = {
            "standard_name": "spectral_response_function",
            "long_name": "spectral response function",
            "units": "",
        }
        logger.debug("add_spectral_response_function: end")

        return result



# Basic pipeline infrastructure
# Aggregate steps
from ._aggregate import AggregateCKDQuad as AggregateCKDQuad
from ._aggregate import AggregateRadiosity as AggregateRadiosity

# Assemble steps
from ._assemble import AddIllumination as AddIllumination
from ._assemble import AddSpectralResponseFunction as AddSpectralResponseFunction
from ._assemble import AddViewingAngles as AddViewingAngles

# Compute steps
from ._compute import ApplySpectralResponseFunction as ApplySpectralResponseFunction
from ._compute import ComputeAlbedo as ComputeAlbedo
from ._compute import ComputeReflectance as ComputeReflectance

from ._core import ApplyCallable as ApplyCallable  # isort: skip
from ._core import Pipeline as Pipeline  # isort: skip
from ._core import PipelineStep as PipelineStep  # isort: skip

# Gather step
from ._gather import GatherCKD as GatherCKD
from ._gather import GatherMono as GatherMono
